# -*- coding: utf-8 -*-

import math
import numpy as np
import pandas as pd
from utils_hoo.utils_general import isnull
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from scipy.stats import norm, lognorm, weibull_min, kstest
from scipy.stats import t, f, chi2
from sklearn.linear_model import LinearRegression as lr
from sklearn.decomposition import PCA
from utils_hoo.Sorts.insert_sort import rank_of_insert
from utils_hoo.Sorts.insert_sort_bin import rank_of_insert_bin


def get_pca(df, dfs_trans=None, **kwargs_PCA):
    '''PCA主成分转化'''
    mdl = PCA(**kwargs_PCA)
    mdl.fit(df)
    def pca_trans(df):
        df_ = mdl.transform(df)
        df_ = pd.DataFrame(df_)
        df_.columns = ['imf'+str(_) for _ in range(1, df_.shape[1]+1)]
        return df_
    df_pca = pca_trans(df)
    if dfs_trans is not None:
        dfs_transed = []
        for df_ in dfs_trans:
            dfs_transed.append(df_)
    else:
        dfs_transed = None
    return df_pca, tuple(dfs_transed)


def fPdf(x, df1, df2, loc=0, scale=1):
    '''自由度为df1和df2的F分布概率密度函数'''
    p = f(df1, df2, loc=loc, scale=scale).pdf(x)
    return p


def fCdf(x, df1, df2, loc=0, scale=1):
    '''自由度为df1和df2的F分布概率密度函数'''
    p = f(df1, df2, loc=loc, scale=scale).cdf(x)
    return p


def fFit(series):
    '''对series（pd.Series或np.array）进行F分布参数估计'''
    df1, df2, loc, scale = f.fit(series)
    return df1, df2, loc, scale


def Fit_f_Pdf(series, x):
    '''
    用series（pd.Series或np.array）拟合一个F分布，并计算在该分布下取值为x的概率
    '''
    df1, df2, loc, scale = fFit(series)
    return fPdf(x, df1, df2, loc, scale)


def Fit_f_Cdf(series, x):
    '''
    用series（pd.Series或np.array）拟合一个F分布，并计算在该分布下取值为x的分布函数值
    '''
    df1, df2, loc, scale = fFit(series)
    return fCdf(x, df1, df2, loc, scale)


def fTest(series, plot=False, **kwargs_plot):
    '''
    对series（pd.Series或np.array）进行F分布检验（目前采用KS检验）
    '''
    df1, df2, loc, scale = f.fit(series)
    Stat, P = kstest(series, 'f', args=(df1, df2, loc, scale))
    if plot:
        from utils_hoo.utils_plot.plot_HistDist import plot_HistDist
        plot_HistDist(series, dists={'f': ('-r', None)}, **kwargs_plot)
    return (Stat, P), (df1, df2, loc, scale)


def tPdf(x, df, loc=0, scale=1):
    '''自由度为df的T分布概率密度函数'''
    # if loc == 0 and scale == 1:
    #     p = math.gamma((df+1)/2) / (math.gamma(df/2) * math.sqrt(df*np.pi)) * \
    #         math.pow(1+math.pow(x, 2)/df, -(df+1)/2)
    p = t(df, loc=loc, scale=scale).pdf(x)
    return p


def tCdf(x, df, loc=0, scale=1):
    '''自由度为df的T分布的分布函数'''
    p = t(df, loc=loc, scale=scale).cdf(x)
    return p


def tFit(series):
    '''对series（pd.Series或np.array）进行T分布参数估计'''
    df, loc, scale = t.fit(series)
    return df, loc, scale


def Fit_t_Pdf(series, x):
    '''
    用series（pd.Series或np.array）拟合一个t分布，并计算在该分布下取值为x的概率
    '''
    df, loc, scale = tFit(series)
    return tPdf(x, df, loc, scale)
    
    
def Fit_t_Cdf(series, x):
    '''
    用series（pd.Series或np.array）拟合一个t分布，并计算在该分布下取值为x的分布函数值
    '''
    df, loc, scale = tFit(series)
    return tCdf(x, df, loc, scale)


def tTest(series, plot=False, **kwargs_plot):
    '''
    对series（pd.Series或np.array）进行t分布检验（目前采用KS检验）
    '''
    df, loc, scale = t.fit(series)
    Stat, P = kstest(series, 't', args=(df, loc, scale))
    if plot:
        from utils_hoo.utils_plot.plot_HistDist import plot_HistDist
        plot_HistDist(series, dists={'t': ('-r', None)}, **kwargs_plot)
    return (Stat, P), (df, loc, scale)


def chi2Pdf(x, df, loc=0, scale=1):
    '''自由度为df的卡方分布概率密度函数'''
    p = chi2(df, loc=loc, scale=scale).pdf(x)
    return p


def chi2Cdf(x, df, loc=0, scale=1):
    '''自由度为df的卡方分布的分布函数'''
    p = chi2(df, loc=loc, scale=scale).cdf(x)
    return p


def chi2Fit(series):
    '''对series（pd.Series或np.array）进行卡方分布参数估计'''
    df, loc, scale = chi2.fit(series)
    return df, loc, scale


def Fit_chi2_Pdf(series, x):
    '''
    用series（pd.Series或np.array）拟合一个卡方分布，并计算在该分布下取值为x的概率
    '''
    df, loc, scale = chi2Fit(series)
    return chi2Pdf(x, df, loc, scale)


def Fit_chi2_Cdf(series, x):
    '''
    用series（pd.Series或np.array）拟合一个卡方分布，并计算在该分布下取值为x的分布函数值
    '''
    df, loc, scale = chi2Fit(series)
    return chi2Cdf(x, df, loc, scale)


def chi2Test(series, plot=False, **kwargs_plot):
    '''
    对series（pd.Series或np.array）进行卡方分布检验（目前采用KS检验）
    '''
    df, loc, scale = chi2.fit(series)
    Stat, P = kstest(series, 'chi2', args=(df, loc, scale))
    if plot:
        from utils_hoo.utils_plot.plot_HistDist import plot_HistDist
        plot_HistDist(series, dists={'chi2': ('-r', None)}, **kwargs_plot)
    return (Stat, P), (df, loc, scale)


def normPdf(mu, sigma, x):
    '''正态分布概率密度函数'''
    p = norm(mu, sigma).pdf(x)
    # p = np.exp(-((x - mu)**2)/(2*sigma**2)) / (sigma * np.sqrt(2*np.pi))
    return p


def normCdf(mu, sigma, x):
    '''正态分布分布函数'''
    p = norm(mu, sigma).cdf(x)
    return p


def normFit(series):
    '''对series（pd.Series或np.array）进行正态分布参数估计'''    
    # mu = series.mean()
    # sigma = series.std(ddof=0)
    mu, sigma = norm.fit(series)
    return mu, sigma


def Fit_norm_Pdf(series, x):
    '''
    用series（pd.Series或np.array）拟合一个正态分布，并计算在该分布下取值为x的概率
    '''
    mu, sigma = normFit(series)
    return normPdf(mu, sigma, x)


def Fit_norm_Cdf(series, x):
    '''
    用series（pd.Series或np.array）拟合一个正态分布，并计算在该分布下取值为x的分布函数值
    '''
    mu, sigma = normFit(series)
    return normCdf(mu, sigma, x)


def normTest(series, plot=False, **kwargs_plot):
    '''
    对series（pd.Series或np.array）进行正态分布检验（目前采用KS检验）
    '''    
    mu, sigma = normFit(series)
    Stat, P = kstest(series, 'norm', args=(mu, sigma))
    if plot:
        from utils_hoo.utils_plot.plot_HistDist import plot_HistDist
        plot_HistDist(series, dists={'norm': ('-r', None)}, **kwargs_plot)
    return (Stat, P), (mu, sigma)


def lognormPdf(mu, sigma, x):
    '''对数正态分布概率密度函数'''
    p = lognorm(s=sigma, loc=0, scale=np.exp(mu)).pdf(x)
    # p = np.exp(-((np.log(x) - mu)**2)/(2 * sigma**2)) / \
    #                                           (x * sigma * np.sqrt(2*np.pi))
    return p


def lognormCdf(mu, sigma, x):
    '''对数正态分布分布函数'''
    p = lognorm(s=sigma, loc=0, scale=np.exp(mu)).cdf(x)
    return p


def lognormFit(series):
    '''对series（pd.Series或np.array）进行对数正态分布参数估计'''
    # stats中lognorm分布参数估计和lognormPdf中参数关系：
    # 若设置floc=0（即始终loc=0），则有s = sigma，scale = e ^ mu
    
    # mu = np.log(series).mean()
    # sigma = np.log(series).std(ddof=0)
    
    s, loc, scale = lognorm.fit(series, floc=0)   
    sigma, mu = s, np.log(scale) 
    
    return mu, sigma


def Fit_lognorm_Pdf(series, x):
    '''
    用series（pd.Series或np.array）拟合一个对数正态分布，并计算在该分布下取值为x的概率
    '''
    mu, sigma = lognormFit(series)
    return lognormPdf(mu, sigma, x)


def Fit_lognorm_Cdf(series, x):
    '''
    用series（pd.Series或np.array）拟合一个对数正态分布，并计算在该分布下取值为x分布函数值
    '''
    mu, sigma = lognormFit(series)
    return lognormCdf(mu, sigma, x)


def lognormTest(series, plot=False, **kwargs_plot):
    '''
    对series（pd.Series或np.array）进行对数正态分布检验（目前采用KS检验）
    '''
    mu, sigma = lognormFit(series)
    s, loc, scale = sigma, 0, np.exp(mu)
    Stat, P = kstest(series, 'lognorm', args=(s, loc, scale))
    if plot:
        from utils_hoo.utils_plot.plot_HistDist import plot_HistDist
        plot_HistDist(series, dists={'lognorm': ('-r', None)}, **kwargs_plot)
    return (Stat, P), (mu, sigma)


def weibullPdf(k, lmd, x):
    '''威布尔分布概率密度函数'''
    p = weibull_min(c=k, loc=0, scale=lmd).pdf(x)
    # p = (k/lmd) * (x/lmd)**(k-1) * np.exp(-(x/lmd)**k)
    return p


def weibullCdf(k, lmd, x):
    '''威布尔分布分布函数'''
    p = weibull_min(c=k, loc=0, scale=lmd).cdf(x)
    return p


def weibullFit(series):
    '''对series（pd.Series或np.array）进行威布尔分布参数估计'''
    # stats中weibull_min分布参数估计和weibullPdf中weibull分布参数关系：
    # 若设置floc=0（即始终loc=0），则有c = k，scale = lmd
    # stats中weibull_min分布参数估计和np.random.weibull分布参数关系：
    # 若设置floc=0（即始终loc=0），则有c = a，scale = 1
    
    c, loc, scale = weibull_min.fit(series, floc=0)
    k, lmd = c, scale
    
    return k, lmd


def Fit_weibull_Pdf(series, x):
    '''
    用series（pd.Series或np.array）拟合一个威布尔分布，并计算在该分布下取值为x的概率
    '''
    k, lmd = weibullFit(series)
    return weibullPdf(k, lmd, x)


def Fit_weibull_Cdf(series, x):
    '''
    用series（pd.Series或np.array）拟合一个威布尔分布，并计算在该分布下取值为x的分布函数值
    '''
    k, lmd = weibullFit(series)
    return weibullCdf(k, lmd, x)


def weibullTest(series, plot=False, **kwargs_plot):
    '''
    对series（pd.Series或np.array）进行威布尔分布检验（目前采用KS检验）
    '''
    k, lmd = weibullFit(series)
    c, loc, scale = k, 0, lmd
    Stat, P = kstest(np.log(series), 'weibull_min', args=(c, loc, scale))
    if plot:
        from utils_hoo.utils_plot.plot_HistDist import plot_HistDist
        plot_HistDist(series, dists={'weibull': ('-r', None)}, **kwargs_plot)
    return (Stat, P), (k, lmd)


def scale_skl(df_fit, dfs_trans=None, cols=None,
              scale_type='std', **kwargs):
    '''
    sklearn数据标准化处理，对df_fit中指定列cols进行标准化
    scale_type指定标准化类型具体如下：
        'std'或'z-score'使用StandardScaler
        'maxmin'或'minmax'使用MinMaxScaler
    **kwargs接收对应Scaler支持的参数
    dfs_trans为需要以df_fit为基准进行标准化的pd.DataFrame列表
    返回标准化后的数据以及sklearn的Scaler对象    
    '''
    
    sklScaler_map = {'std': StandardScaler, 'z-score': StandardScaler,
                     'maxmin': MinMaxScaler, 'minmax': MinMaxScaler}
    sklScaler = sklScaler_map[scale_type]
    
    cols_all = list(df_fit.columns)
    
    if cols is None:
        scaler = sklScaler(**kwargs).fit(df_fit)
        df_fited = pd.DataFrame(scaler.transform(df_fit),
                                         columns=cols_all, index=df_fit.index)
        if dfs_trans is None:
            return df_fited, None, (scaler, cols_all)
        dfs_transed = [pd.DataFrame(scaler.transform(df),
                       columns=cols_all, index=df.index) for df in dfs_trans]
        return df_fited, dfs_transed, (scaler, df_fit.columns)
    
    cols_rest = [x for x in cols_all if x not in cols]    
    df_toFit = df_fit.reindex(columns=cols)
    
    scaler = sklScaler(**kwargs).fit(df_toFit)   
    df_fited = pd.DataFrame(scaler.transform(df_toFit), columns=cols,
                                                        index=df_toFit.index)
    for col in cols_rest:
        df_fited[col] = df_fit[col]
    df_fited = df_fited.reindex(columns=cols_all)      
    if dfs_trans is None:
        return df_fited, None, (scaler, cols)    
    dfs_transed = []
    for df in dfs_trans:
        df_trans = df.reindex(columns=cols)
        df_transed = pd.DataFrame(scaler.transform(df_trans),
                                  columns=cols, index=df_trans.index)
        cols_all_df = list(df.columns)
        cols_rest_df = [x for x in cols_all_df if x not in cols]
        for col in cols_rest_df:
            df_transed[col] = df[col]
        dfs_transed.append(df_transed.reindex(columns=cols_all_df))        
    return df_fited, tuple(dfs_transed), (scaler, cols)


def scale_skl_inverse(scaler, dfs_toInv, cols=None, **kwargs):
    '''
    反标准化还原数据
    scaler为fit过的sklearn Scaler类(如StandardScaler、MinMaxScaler)
    dfs_toInv为待还原的df列表
    **kwargs接收scaler.inverse_transform函数支持的参数
    注：(scaler, cols)应与scale_skl函数输出一致
    '''
    
    dfs_inved = []
    
    if cols is None:
        for df in dfs_toInv:
            df_inved = pd.DataFrame(scaler.inverse_transform(df, **kwargs), 
                                    columns=df.columns, index=df.index)
            dfs_inved.append(df_inved)
        return tuple(dfs_inved)
    
    for df in dfs_toInv:
        cols_all = list(df.columns)
        cols_rest = [x for x in cols_all if x not in cols]
        df_inved = pd.DataFrame(scaler.inverse_transform(df[cols], **kwargs),
                                columns=cols, index=df.index)
        for col in cols_rest:
            df_inved[col] = df[col]
        dfs_inved.append(df_inved.reindex(columns=cols_all))
        
    return tuple(dfs_inved)


def norm_std(series, isReverse=False, ddof=1, returnMeanStd=False):
    '''
    z-score标准化，series为pd.Series或np.array，isReverse设置是否反向
    ddof指定计算标准差时是无偏还是有偏的：
        ddof=1时计算的是无偏标准差（样本标准差，分母为n-1），
        ddof=0时计算的是有偏标准差（总体标准差，分母为n）
    （注：pandas的std()默认计算的是无偏标准差，numpy的std()默认计算的是有偏标准差）
    当returnMeanStd为True时同时返回均值和标准差，为False时不返回
    
    注: Z-score适用于series的最大值和最小值未知或有超出取值范围的离群值的情况。
        （一般要求原始数据的分布可以近似为高斯分布，否则效果可能会很糟糕）
    总体标准差和样本标准差参考:
        https://blog.csdn.net/qxqxqzzz/article/details/88663198
    '''
    Smean, Sstd = series.mean(), series.std(ddof=ddof)
    if not isReverse:
        series_new = (series - Smean) / Sstd
    else:
        series_new = (Smean - series) / Sstd
    if not returnMeanStd:
        return series_new
    else:
        return series_new, (Smean, Sstd)


def norm_linear(x, Xmin, Xmax, Nmin=0, Nmax=1, isReverse=False,
                xMustInRange=True, vXminEqXmax=np.nan, vNminEqNmax=np.nan):
    '''
    线性映射：将取值范围在[Xmin, Xmax]内的x映射到取值范围在[Nmin, Nmax]内的xNew
    isReverse设置是否反向，若为True，则映射到[Nmax, Nmin]范围内
    xMustInRange设置当x不在[Xmin, Xmax]范围内时是否报错
    vXminEqXmax设置当Xmin和Xmax相等时的返回值
    '''
    
    if xMustInRange:
        if x < Xmin or x > Xmax:
            raise ValueError('必须满足 Xmin =< x <= Xmax ！')
    
    if Xmin > Xmax or Nmin > Nmax:
        raise ValueError('必须满足 Xmin <= Xmax 且 Nmin <= Nmax ！')
        
    if Xmin == Xmax:
        return vXminEqXmax
    
    if Nmin == Nmax:
        return vNminEqNmax
    
    if isReverse:
        Nmin, Nmax = Nmax, Nmin
    
    xNew = Nmin + (x-Xmin) * (Nmax-Nmin) / (Xmax-Xmin)
    
    return xNew


def norm_mid(x, x_min, x_max, Nmin=0, Nmax=1, x_best=None):
    '''
    中间型（倒V型）指标的正向化线性映射，新值范围为[Nmin, Nmax]
    (指标值既不要太大也不要太小，适当取中间值最好，如水质量评估PH值)
    x_min和x_max为指标可能取到的最小值和最大值
    x_best指定最优值，若不指定则将x_min和x_max的均值当成最优值
    参考：https://zhuanlan.zhihu.com/p/37738503
    '''    
    x_best = (x_min+x_max)/2 if x_best is None else x_best
    if x <= x_min or x >= x_max:
        return Nmin
    elif x > x_min and x < x_best:
        return norm_linear(x, x_min, x_best, Nmin, Nmax)
    elif x < x_max and x >= x_best:
        return norm_linear(x, x_best, x_max, Nmin, Nmax, isReverse=True)


def norm01_mid(x, x_min, x_max, x_best=None):
    '''
    中间型（倒V型）指标的正向化和（线性）01标准化
    (指标值既不要太大也不要太小，适当取中间值最好，如水质量评估PH值)
    x_min和x_max为指标可能取到的最小值和最大值
    x_best指定最优值，若不指定则将x_min和x_max的均值当成最优值
    参考：https://zhuanlan.zhihu.com/p/37738503
    '''    
    x_best = (x_min+x_max)/2 if x_best is None else x_best
    if x <= x_min or x >= x_max:
        return 0
    elif x > x_min and x < x_best:
        return (x - x_min) / (x_best - x_min)    
    elif x < x_max and x >= x_best:
        return (x_max - x) / (x_max - x_best)


def norm_side(x, x_min, x_max, Nmin=0, Nmax=1, x_worst=None, v_outLimit=None):
    '''
    两边型（V型）指标的正向化线性映射，新值范围为[Nmin, Nmax]
    (指标越靠近x_min或越靠近x_max越好，越在中间越不好)
    x_min和x_max为指标可能取到的最小值和最大值
    x_worst指定最差值，若不指定则将x_min和x_max的均值当成最差值
    v_outLimit指定当x超过x_min或x_max界限之后的正向标准化值，不指定则默认为Nmax
    '''
    v_outLimit = Nmax if v_outLimit is None else v_outLimit
    x_worst = (x_min+x_max)/2 if x_worst is None else x_worst
    if x < x_min or x > x_max:
        return v_outLimit
    elif x >= x_min and x < x_worst:
        return norm_linear(x, x_min, x_worst, Nmin, Nmax, isReverse=True)
    elif x <= x_max and x >= x_worst:
        return norm_linear(x, x_worst, x_max, Nmin, Nmax)
    
    
def norm01_side(x, x_min, x_max, x_worst=None, v_outLimit=1):
    '''
    两边型（V型）指标的正向化和（线性）01标准化
    (指标越靠近x_min或越靠近x_max越好，越在中间越不好)
    x_min和x_max为指标可能取到的最小值和最大值
    x_worst指定最差值，若不指定则将x_min和x_max的均值当成最差值
    v_outLimit指定当x超过x_min或x_max界限之后的正向标准化值，不指定则默认为1
    '''
    x_worst = (x_min+x_max)/2 if x_worst is None else x_worst
    if x < x_min or x > x_max:
        return v_outLimit
    elif x >= x_min and x < x_worst:
        return (x_worst - x) / (x_worst - x_min) 
    elif x <= x_max and x >= x_worst:
        return (x - x_worst) / (x_max - x_worst)
    
    
def norm_range(x, x_min, x_max, x_minMin, x_maxMax, Nmin=0, Nmax=1):
    '''
    区间型指标的正向化正向化线性映射，新值范围为[Nmin, Nmax]
    (指标的取值最好落在某一个确定的区间最好，如体温)
    [x_min, x_max]指定指标的最佳稳定区间，[x_minMin, x_maxMax]指定指标的最大容忍区间
    参考：https://zhuanlan.zhihu.com/p/37738503
    '''
    if x >= x_min and x <= x_max:
        return Nmax
    elif x <= x_minMin or x >= x_maxMax:
        return Nmin
    elif x > x_max and x < x_maxMax:
        return norm_linear(x, x_max, x_maxMax, Nmin, Nmax, isReverse=True)
    elif x < x_min and x > x_minMin:
        return norm_linear(x, x_minMin, x_min, Nmin, Nmax)


def norm01_range(x, x_min, x_max, x_minMin, x_maxMax):
    '''
    区间型指标的正向化和（线性）01标准化
    (指标的取值最好落在某一个确定的区间最好，如体温)
    [x_min, x_max]指定指标的最佳稳定区间，[x_minMin, x_maxMax]指定指标的最大容忍区间
    参考：https://zhuanlan.zhihu.com/p/37738503
    '''
    if x >= x_min and x <= x_max:
        return 1
    elif x <= x_minMin or x >= x_maxMax:
        return 0
    elif x > x_max and x < x_maxMax:
        return 1 - (x - x_max) / (x_maxMax - x_max)
    elif x < x_min and x > x_minMin:
        return 1 - (x_min - x) / (x_min - x_minMin)
        
        
def mse(y_true, y_predict):
    '''
    MSE，y_true和y_predict格式为np.array或pd.Series
    注: 要求y_true和y_predict值一一对应
    '''
    if y_predict.shape != y_true.shape:
        raise ValueError('y_true和y_predict的值必须一一对应！')
    yTrue = pd.Series(y_true).reset_index(drop=True)
    yPred = pd.Series(y_predict).reset_index(drop=True)
    return ((yTrue - yPred) ** 2).mean()


def r2(y_true, y_predict, is_linear=False):
    '''
    计算拟合优度R2
    在线性回归情况下（y_predict是由y_true与自变量X进行线性拟合的预测值），有:
        R2 = 1 - SSres / SStot = SSreg / SStot，
        且此时R2与y_true和y_predict相关系数的平方相等
    非线性回归情况下，1 - SSres / SStot != SSreg / SStot，R2与两者相关系数平方也不相等
    可设置is_linear为True和False进行比较验证
    参考：https://blog.csdn.net/wade1203/article/details/98477034
         https://blog.csdn.net/liangyihuai/article/details/88560859
         https://wenku.baidu.com/view/893b22d66bec0975f465e2b8.html
    '''
    if y_predict.shape != y_true.shape:
        raise ValueError('y_true和y_predict的值必须一一对应！')
    yTrue = pd.Series(y_true).reset_index(drop=True)
    yPred = pd.Series(y_predict).reset_index(drop=True)    
    SStot = sum((yTrue - yTrue.mean()) ** 2)
    if not is_linear:
        SSres = sum((yTrue - yPred) ** 2)
        return 1 - SSres / SStot
    else:
        SSreg = sum((yPred - yTrue.mean()) ** 2)
        return SSreg / SStot
    
    
def r2adj(y_true, y_predict, k):
    '''
    计算调整R2, k为自变量个数
    注：线性回归中常量也算一个变量，例如y=ax+b，则k为2
    http://blog.sciencenet.cn/blog-651374-975670.html
    '''
    vR2 = r2(y_true, y_predict)
    n = len(y_true)
    return 1 - (1-vR2) * ((n-1) / (n-k))


def r2adj_by_r2(vR2, n, k):
    '''
    通过R2计算调整R2, n为样本量, k为自变量个数
    注：线性回归中常量也算一个变量，例如y=ax+b，则k为2
    '''
    return 1 - (1-vR2) * ((n-1) / (n-k))


def r2adj_by_mse(vMSE, y_true, k):
    '''
    通过MSE计算调整R2, n为样本量, k为自变量个数
    注：线性回归中常量也算一个变量，例如y=ax+b，则k为2
    '''
    n = len(y_true)
    vR2 = r2_by_mse(vMSE, y_true)
    return 1 - (1-vR2) * ((n-1) / (n-k))


def r2_by_r2adj(vR2adj, n, k):
    '''
    通过调整R2计算R2, n为样本量, k为自变量个数
    注：线性回归中常量也算一个变量，例如y=ax+b，则k为2    
    '''
    return 1 - (1-vR2adj) * ((n-k) / (n-1))
    
    
def r2_by_mse(vMSE, y_true):
    '''
    根据MSE和真实值计算拟合优度R2
    参考：https://blog.csdn.net/wade1203/article/details/98477034
         https://blog.csdn.net/liangyihuai/article/details/88560859
    '''
    y_true_var = pd.Series(y_true).var(ddof=0)
    return 1 - vMSE / y_true_var


def mse_by_r2(vR2, y_true):
    '''
    根据R2和真实值计算MSE
    '''
    y_true_var = pd.Series(y_true).var(ddof=0)
    return (1-vR2) * y_true_var


def r2_deprecated(y_true, y_predict):
    '''
    拟合优度R2计算
    参考：https://blog.csdn.net/wade1203/article/details/98477034
         https://blog.csdn.net/liangyihuai/article/details/88560859
    '''
    vMSE = mse(y_true, y_predict)
    y_true_var = pd.Series(y_true).var(ddof=0)
    return 1 - vMSE / y_true_var
    
    
def rmse(y_true, y_predict):
    '''
    RMSE，y_true和y_predict格式为np.array或pd.Series
    注: 要求y_true和y_predict值一一对应
    '''
    return mse(y_true, y_predict) ** 0.5
    
    
def mae(y_true, y_predict):
    '''
    MAE，y_true和y_predict格式为np.array或pd.Series
    注: 要求y_true和y_predict值一一对应
    '''
    if y_predict.shape != y_true.shape:
        raise ValueError('y_true和y_predict的值必须一一对应！')
    yTrue = pd.Series(y_true).reset_index(drop=True)
    yPred = pd.Series(y_predict).reset_index(drop=True)
    return (abs(yTrue-yPred)).mean()
    
    
def mape(y_true, y_predict):
    '''
    MAPE，y_true和y_predict格式为np.array或pd.Series
    注: 要求y_true和y_predict值一一对应
    注：当y_true存在0值时MAPE不可用
    '''
    if y_predict.shape != y_true.shape:
        raise ValueError('y_true和y_predict的值必须一一对应！')
    yTrue = pd.Series(y_true).reset_index(drop=True)
    yPred = pd.Series(y_predict).reset_index(drop=True)
    return (abs((yPred-yTrue) / yTrue)).mean()
        
        
def smape(y_true, y_predict):
    '''
    SMAPE，y_true和y_predict格式为np.array或pd.Series
    注: 要求y_true和y_predict值一一对应
    注：当y_true和y_predict均为0值时SMAPE不可用
    公式：https://blog.csdn.net/guolindonggld/article/details/87856780
    '''
    if y_predict.shape != y_true.shape:
        raise ValueError('y_true和y_predict的值必须一一对应！')
    yTrue = pd.Series(y_true).reset_index(drop=True)
    yPred = pd.Series(y_predict).reset_index(drop=True)
    return (abs(yPred-yTrue) / ((abs(yPred)+abs(yTrue)) / 2)).mean()


def AVEDEV(arr):
    '''
    AVEDEV函数
    arr为np.array或pd.Series
    
    参考：
    https://support.microsoft.com/en-us/office/
    avedev-function-58fe8d65-2a84-4dc7-8052-f3f87b5c6639?ui=en-us&
    rs=en-us&ad=us
    '''
    
    return (abs(arr-arr.mean())).mean()


def cal_linear_reg_r(y, x=None):
    '''
    计算y中数据点的斜率（一元线性回归）
    y和x为list或pd.Series或np.array
    '''
    if isnull(x):
        X = pd.DataFrame({'X': range(0, len(y))})
    else:
        X = pd.DataFrame({'X': x})
    y = pd.Series(y)
    mdl = lr().fit(X, y)
    return mdl.coef_[0], mdl.intercept_


def cummean(series, skipna=True):
    '''
    滚动计算序列累计均值
    skip设置是否忽略(删除)无效值记录
    '''
    df = pd.DataFrame({'series': series})
    df['cumsum'] = df['series'].cumsum(skipna=skipna)
    df['N'] = range(1, df.shape[0]+1)
    if skipna:
        df['N'] = df['series'].notna().cumsum()
    df['cummean'] = df['cumsum'] / df['N']
    return df['cummean']


def delta_var_d0(n0, mean0, var0, n1, mean1, var1):
    '''
    自由度为0情况下的增量方差算法
    n, mean, var分别为两组样本的样本量、均值和自由度为0的方差
    https://www.cnblogs.com/yoyaprogrammer/p/delta_variance.html
    '''
    if n0 < 1 or n1 < 1:
        raise ValueError('样本量必须大于1！')
    fm = n0 + n1
    mean = (n0 * mean0 + n1 * mean1) / fm
    fz1 = n0 * (var0 + (mean - mean0) ** 2)
    fz2 = n1 * (var1 + (mean - mean1) ** 2)
    var = (fz1 + fz2) / fm
    return var


def delta_var_d(n0, mean0, var0, n1, mean1, var1, ddof=1):
    '''
    增量方差算法：根据已经算好的两组样本的方差、均值和样本量计算两组数合并之后的方差
    ddof为自由度
    n, mean, var分别为两组样本的样本量、均值和自由度为ddof的方差
    https://www.cnblogs.com/yoyaprogrammer/p/delta_variance.html
    '''
    fm = n0 + n1
    if n0 <= ddof or n1 <= ddof or fm <= ddof: # 样本量必须大于自由度
        return np.nan
    # 自由度还原为0
    var0_0, var1_0 = ((n0-ddof) / n0) * var0, ((n1-ddof) / n1) * var1
    var_0 = delta_var_d0(n0, mean0, var0_0, n1, mean1, var1_0)
    var = (fm / (fm-ddof)) * var_0 # 自由度转换
    return var


def delta_var(n0, mean0, var0, n1, mean1, var1, ddof=1):
    '''
    增量方差算法
    '''
    n = n0+n1
    if n0 <= ddof or n1 <= ddof or n <= ddof: # 样本量必须大于自由度
        return np.nan
    fm = n - ddof
    mean = (n0 * mean0 + n1 * mean1) / n
    fz1 = (n0-ddof) * var0 + n0 * (mean - mean0) ** 2
    fz2 = (n1-ddof) * var1 + n1 * (mean - mean1) ** 2
    var = (fz1 + fz2) / fm
    return var


def cumvar_nonan(series, ddof=1):
    '''
    滚动累计方差算法
    '''
    
    # 数据格式检查
    series = np.array(series)
    assert len(series.shape) == 1, 'series必须为1维！'
    # assert sum(np.isnan(series)) == 0, 'series中不能有无效值！'
    
    # 累计均值
    cummean = np.cumsum(series) / np.arange(1, len(series)+1)
    
    # 累计方差
    cumvar = np.nan * np.ones(len(series),)
    if ddof == 0:
        cumvar[0] = 0
    else:
        for k in range(ddof, ddof+ddof+1):
            cumvar[k] = np.var(series[:k+1], ddof=ddof)
    for k in range(ddof+ddof+1, len(series)):
        var0, mean0, n0 = cumvar[k-ddof-1], cummean[k-ddof-1], k-ddof
        var1 = np.var(series[k-ddof:k+1], ddof=ddof)
        mean1 = np.mean(series[k-ddof:k+1])
        # 增量方差
        cumvar[k] = delta_var(n0, mean0, var0, ddof+1, mean1, var1, ddof=ddof)
    return cumvar


def cumvar(series, ddof=1, skipna=True):
    '''滚动累计方差算法，返回pd.Series'''
    df = pd.DataFrame({'v': series})
    if not skipna:
        df['cumvar'] = cumvar_nonan(df['v'], ddof=ddof)
        return df['cumvar']
    else:
        df_ = df.dropna(subset=['v'], how='any').copy()
        df_['cumvar'] = cumvar(df_['v'], ddof=ddof, skipna=False)
        df['cumvar'] = df_['cumvar']
        return df['cumvar']
    

def cumvar1(series, ddof=1, skipna=True):
    '''
    滚动计算序列累计方差
    参考：https://www.cnblogs.com/yoyaprogrammer/p/delta_variance.html
    '''
    
    df = pd.DataFrame({'v': series})
    ori_idx = df.index
    df.reset_index(drop=True, inplace=True)
    
    df['cummean'] = cummean(df['v'], skipna=skipna)
    
    if skipna:
        df_var = df.dropna(subset=['v'], how='any').copy()
    else:
        df_var = df.copy()
        
    if ddof == 0:
        df_var['cumvar'] = 0
        for k in range(1, df_var.shape[0]):
            var0 = df_var.loc[df_var.index[k-1], 'cumvar']
            mean0 = df_var.loc[df_var.index[k-1], 'cummean']
            mean1 = df_var.loc[df_var.index[k], 'v']
            var = delta_var(k, mean0, var0, 1, mean1, 0, ddof=0)
            df_var.loc[df_var.index[k], 'cumvar'] = var
    else:        
        df_var['roll_var'] = df_var['v'].rolling(ddof+1).var(skipna=skipna,
                                                             ddof=ddof)
        df_var['roll_mean'] = df_var['v'].rolling(ddof+1).mean(skipna=skipna)
        df_var['cumvar'] = np.nan
        for k in range(ddof, ddof+ddof+1):
            df_var.loc[df_var.index[k], 'cumvar'] = \
                        df_var['v'].iloc[:k+1].var(skipna=skipna, ddof=ddof)
        for k in range(ddof+ddof+1, df_var.shape[0]):
            var0 = df_var.loc[df_var.index[k-ddof-1], 'cumvar']
            mean0 = df_var.loc[df_var.index[k-ddof-1], 'cummean']
            n0 = k - ddof
            var1 = df_var.loc[df_var.index[k], 'roll_var']
            mean1 = df_var.loc[df_var.index[k], 'roll_mean']
            n1 = ddof + 1
            var = delta_var(n0, mean0, var0, n1, mean1, var1, ddof=ddof)
            df_var.loc[df_var.index[k], 'cumvar'] = var
        
    df['cumvar'] = df_var['cumvar']
    
    df.index = ori_idx
    
    return df['cumvar']


def cumvar2(series, ddof=1, skipna=True):
    '''
    滚动计算序列累计方差
    https://www.cnblogs.com/yoyaprogrammer/p/delta_variance.html
    '''
    
    df = pd.DataFrame({'v': series})
    ori_idx = df.index
    df.reset_index(drop=True, inplace=True)
    
    df['cumsum'] = df['v'].cumsum(skipna=skipna)
    df['N'] = range(1, df.shape[0]+1)
    if skipna:
        df['N'] = df['v'].notna().cumsum()
    df['cummean'] = df['cumsum'] / df['N']
    
    if skipna:
        df_var = df.dropna(subset=['v'], how='any').copy()
    else:
        df_var = df.copy()
        
    if ddof == 0:
        df_var['cumvar'] = 0
        for k in range(1, df_var.shape[0]):
            var0 = df_var.loc[df_var.index[k-1], 'cumvar']
            mean0 = df_var.loc[df_var.index[k-1], 'cummean']
            mean1 = df_var.loc[df_var.index[k], 'v']
            var = delta_var_d0(k, mean0, var0, 1, mean1, 0)
            df_var.loc[df_var.index[k], 'cumvar'] = var
    else:        
        df_var['roll_var'] = df_var['v'].rolling(ddof+1).var(skipna=skipna,
                                                             ddof=ddof)
        df_var['roll_mean'] = df_var['v'].rolling(ddof+1).mean(skipna=skipna)
        df_var['cumvar'] = np.nan
        for k in range(ddof, ddof+ddof+1):
            df_var.loc[df_var.index[k], 'cumvar'] = \
                        df_var['v'].iloc[:k+1].var(skipna=skipna, ddof=ddof)
        for k in range(ddof+ddof+1, df_var.shape[0]):
            var0 = df_var.loc[df_var.index[k-ddof-1], 'cumvar']
            mean0 = df_var.loc[df_var.index[k-ddof-1], 'cummean']
            n0 = k - ddof
            var1 = df_var.loc[df_var.index[k], 'roll_var']
            mean1 = df_var.loc[df_var.index[k], 'roll_mean']
            n1 = ddof + 1
            var = delta_var_d(n0, mean0, var0, n1, mean1, var1, ddof=ddof)
            df_var.loc[df_var.index[k], 'cumvar'] = var
        
    df['cumvar'] = df_var['cumvar']
    
    df.index = ori_idx
    
    return df['cumvar']


def cumstd(series, ddof=1, skipna=True):
    '''滚动计算序列累计标准差'''
    return np.sqrt(cumvar(series, ddof=ddof, skipna=skipna))
    
    
def cumvar_Deprecated(series, **kwargs):
    '''
    滚动计算序列累计方差
    '''
    df = pd.DataFrame({'v': series})
    df['cumvar'] = np.nan
    ori_idx = df.index
    df.reset_index(drop=True, inplace=True)
    for k in range(df.shape[0]):
        df.loc[df.index[k], 'cumvar'] = \
                                df.loc[df.index[:k+1], 'v'].var(**kwargs)
    df.index = ori_idx
    return df['cumvar']


def cumstd_Deprecated(series, **kwargs):
    '''
    滚动计算序列累计标准差
    '''
    df = pd.DataFrame({'series': series})
    df['cumstd'] = np.nan
    ori_idx = df.index
    df.reset_index(drop=True, inplace=True)
    for k in range(df.shape[0]):
        df.loc[df.index[k], 'cumstd'] = \
                                df.loc[df.index[:k+1], 'series'].std(**kwargs)
    df.index = ori_idx
    return df['cumstd']


def cumScale(series, scale_type=('maxmin', 0, 1)):
    '''
    序列累计标准化
    scale_type可选：
        最大最小标准化：('maxmin', Nmin, Nmax)
        标准偏差法：('std', ) 或 'std'
    '''
    df = pd.DataFrame({'v': series})
    if scale_type[0] == 'maxmin':
        Nmin, Nmax = scale_type[1], scale_type[2]
        x = df['v']
        Xmin = df['v'].cummin()
        Xmax = df['v'].cummax()
        df['scale'] = Nmin + (x-Xmin) * (Nmax-Nmin) / (Xmax-Xmin)
    elif scale_type in ['std', 'z-score'] or \
                                        scale_type[0] in ['std', 'z-score']:
        df['cummean'] = cummean(series)
        df['cumstd'] = cumstd(series)
        df['scale'] = (df['v'] - df['cummean']) / df['cumstd']
    return df['scale']


def rollingScale(series, lag, scale_type=('maxmin', 0, 1)):
    '''序列滚动标准化'''
    df = pd.DataFrame({'v': series})
    if scale_type[0] == 'maxmin':
        Nmin, Nmax = scale_type[1], scale_type[2]
        x = df['v']
        Xmin = df['v'].rolling(lag).min()
        Xmax = df['v'].rolling(lag).max()
        df['scale'] = Nmin + (x-Xmin) * (Nmax-Nmin) / (Xmax-Xmin)
    elif scale_type in ['std', 'z-score'] or \
                                        scale_type[0] in ['std', 'z-score']:
        df['mean_dy'] = df['v'].rolling(lag).mean()
        df['std_dy'] = df['v'].rolling(lag).std()
        df['scale'] = (df['v'] - df['mean_dy']) / df['std_dy']
    return df['scale']


def get_quantiles(series, method='dense'):
    '''
    计算series（pd.Series或np.array）中每个数所处的百分位
    method决定并列排序序号确定方式
    '''
    if not method in ['dense', 'average']:
        raise ValueError('未识别的并列排序方法！')
    series = pd.Series(series)
    ranks = series.rank(method=method)
    if method == 'average':
        return ranks / len(ranks)
    elif method == 'dense':
        return ranks / ranks.max()
    
    
def cumNunique(series, skipna=True):
    '''滚动计算series中unique值的个数'''    
    vals = list(series)
    n = len(vals)
    if skipna:
        Nunique = [np.nan] * n
        k0 = 0
        while k0 < n and isnull(vals[k0]):
            k0 += 1
        Nunique[k0] = 1
        vunique = [vals[k0]]
        for k in range(k0+1, n):
            if isnull(vals[k]):
                continue
            else:
                if vals[k] in vunique:
                    Nunique[k] = Nunique[k0]
                else:
                    vunique.append(vals[k])
                    Nunique[k] = Nunique[k0] + 1
                k0 += 1
    else:        
        Nunique = [1] * n
        vunique = [vals[0]]
        for k in range(1, n):
            if vals[k] in vunique:
                Nunique[k] = Nunique[k-1]
            else:
                vunique.append(vals[k])
                Nunique[k] = Nunique[k-1] + 1
    return Nunique


def cumNunique_pd(series, skipna=True):
    '''滚动计算series中unique值的个数'''
    
    df = pd.DataFrame({'v': series})
    if skipna:
        df_ = df.dropna(subset=['v'], how='any').copy()
    else:
        df_ = df.copy()
    df_['Nunique'] = ~df_['v'].duplicated()
    df_['Nunique'] = df_['Nunique'].cumsum()
    
    df['Nunique'] = df_['Nunique']
    
    return df['Nunique']


def cumrank_nonan(series, ascending=True, method='dense'):
    '''
    滚动计算累计排序号
    '''
    
    series = list(series)
    n = len(series)
    cumranks = [1] * n
    
    nums_sorted = [series[0]]
    ranks = [1]
    for k in range(1, n):
        num = series[k]
        irank, (nums_sorted, ranks) = rank_of_insert_bin(nums_sorted, ranks,
                                    num, ascending=ascending, method=method)
        cumranks[k] = irank
    
    return cumranks


def cumrank(series, ascending=True, method='dense'):
    '''
    滚动计算累计排序号
    '''
    df = pd.DataFrame({'v': series})
    if df['v'].isnull().sum() == 0:
        return cumrank_nonan(series, ascending=ascending, method=method)
    else:
        df_ = df.copy()
        df_.dropna(subset=['v'], how='any', inplace=True)
        df_['cumrank'] = cumrank_nonan(df_['v'], ascending=ascending, method=method)
        df['cumrank'] = df_['cumrank']
        return df['cumrank']


def cumrank1_nonan(series, ascending=True, method='dense'):
    '''
    滚动计算累计排序号
    '''
    
    series = list(series)
    n = len(series)
    cumranks = [1] * n
    
    nums_sorted = [series[0]]
    ranks = [1]
    for k in range(1, n):
        num = series[k]
        irank, (nums_sorted, ranks) = rank_of_insert(nums_sorted, ranks,
                                    num, ascending=ascending, method=method)
        cumranks[k] = irank
    
    return cumranks


def cumrank1(series, ascending=True, method='dense'):
    '''
    滚动计算累计排序号
    '''
    df = pd.DataFrame({'v': series})
    if df['v'].isnull().sum() == 0:
        return cumrank1_nonan(series, ascending=ascending, method=method)
    else:
        df_ = df.copy()
        df_.dropna(subset=['v'], how='any', inplace=True)
        df_['cumrank'] = cumrank1_nonan(df_['v'], ascending=ascending, method=method)
        df['cumrank'] = df_['cumrank']
        return df['cumrank']
    
    
def cumrank2(series, ascending=True, method='dense'):
    '''
    滚动计算累计排序号
    '''
    df = pd.DataFrame({'v': series})
    ori_idx = df.index
    df.reset_index(drop=True, inplace=True)
    df['cumrank'] = np.nan
    for k in range(0, df.shape[0]):
        df.loc[df.index[k], 'cumrank'] = df.loc[df.index[:k+1], 'v'].rank(
                                  ascending=ascending, method=method).iloc[-1]
    df.index = ori_idx
    return df['cumrank']


def get_pct_loc(value, series, isnew=True, method='dense'):
    '''
    给定value和series（pd.Series或np.array）计算value在series中所处百分位
    若isnew为True，则将value当成新数据，若为False，则当成series中已经存在的数据
    method决定并列排序序号确定方式
    '''
    if isnull(value):
        return np.nan
    vals = [value] + list(series)
    vals = pd.Series(vals)
    if not method in ['dense', 'average']:
        raise ValueError('未识别的并列排序方法！')
    ranks = vals.rank(method=method)
    rank = ranks.iloc[0]
    if method == 'dense':
        return rank / ranks.max()
    elif method == 'average':
        if isnew or list(ranks).count(rank) == 1:
            return rank / len(ranks)
        else:
            return (rank-0.5) / (len(vals)-1)
        
        
def cumPctLoc(series, method='dense'):
    '''滚动计算数据累计所处百分位'''
    if not method in ['dense', 'average']:
        raise ValueError('未识别的并列排序方法！')
    df = pd.DataFrame({'v': series})
    df['cumrank'] = cumrank(df['v'], method=method)
    if method == 'dense':
        df['cumNunique'] = cumNunique_pd(df['v'])
        df['PctLoc'] = df['cumrank'] / df['cumNunique']
    elif method == 'average':
        df['idx'] = range(1, df.shape[0]+1)
        df['PctLoc'] = df['cumrank'] / df['idx']
    return df['PctLoc']
        
        
def cumPctLoc1(series, method='dense'):
    '''滚动计算数据累计所处百分位'''
    if not method in ['dense', 'average']:
        raise ValueError('未识别的并列排序方法！')
    df = pd.DataFrame({'v': series})
    df['cumrank'] = cumrank(df['v'], method=method)
    if method == 'dense':
        df['cumNunique'] = cumNunique(df['v'])
        df['PctLoc'] = df['cumrank'] / df['cumNunique']
    elif method == 'average':
        df['idx'] = range(1, df.shape[0]+1)
        df['PctLoc'] = df['cumrank'] / df['idx']
    return df['PctLoc']
        
    
def cumPctLoc2(series, method='dense'):
    '''滚动计算数据累计所处百分位'''
    df = pd.DataFrame({'v': series})
    df['PctLoc'] = np.nan
    ori_idx = df.index
    df.reset_index(drop=True, inplace=True)
    df.loc[df.index[0], 'PctLoc'] = 1.0
    if isnull(df.loc[df.index[0], 'v']):
        df.loc[df.index[0], 'PctLoc'] = np.nan
    for k in range(1, df.shape[0]):
        value = df.loc[df.index[k], 'v']
        values = df.loc[df.index[:k], 'v']
        df.loc[df.index[k], 'PctLoc'] = get_pct_loc(value, values,
                                                    isnew=True, method=method)
    df.index = ori_idx
    return df['PctLoc']


def rollingPctLoc(series, lag, method='dense'):
    '''滚动计算数据所处百分位'''
    df = pd.DataFrame({'v': series})
    ori_idx = df.index
    df.reset_index(drop=True, inplace=True)
    # df['PctLoc'] = np.nan
    # for k in range(0, df.shape[0]-lag+1):
    #     value = df.loc[df.index[k+lag-1], 'v']
    #     values = df.loc[df.index[k:k+lag-1], 'v']
    #     df.loc[df.index[k+lag-1], 'PctLoc'] = \
    #                     get_pct_loc(value, values, isnew=True, method=method)
    df['PctLoc'] = df['v'].rolling(lag).apply(lambda x: 
              get_pct_loc(x.iloc[-1], x.iloc[:-1], isnew=True, method=method))
    df.index = ori_idx
    return df['PctLoc']


class External_Std(object):
    '''极端值处理，标准差倍数法'''
    
    def __init__(self, nstd=3, cols=None):
        raise NotImplementedError
    
    # def deal_ext_value(df_fit, cols=None, dfs_trans=None, nstd=5):
    #     '''极端值处理'''
        
    #     cols = df_fit.columns if cols is None else cols
        
    #     mean_stds = []
    #     for col in cols:
    #         mean_stds.append((col, df_fit[col].mean(), df_fit[col].std()))
            
    #     df_fited = df_fit.copy()
    #     for (col, mean, std) in mean_stds:
    #         df_fited[col] = df_fited[col].apply(
    #                         lambda x: np.clip(x, mean-nstd*std, mean+nstd*std))   
    
    #     if dfs_trans is not None:
    #         dfs_traned = []
    #         for df in dfs_trans:
    #             df_tmp = df.copy()
    #             for (col, mean, std) in mean_stds:
    #                 df_tmp[col] = df_tmp[col].apply(
    #                         lambda x: np.clip(x, mean-nstd*std, mean+nstd*std))
    #             dfs_traned.append(df_tmp)
        
    #     return df_fited, tuple(dfs_traned)


def parmsEst():
    '''各种分布不同方法的参数估计'''
    raise NotImplementedError
    
    
def auc():
    raise NotImplementedError


def VarHomTest(s1, s2):
    '''方差齐性检验'''
    raise NotImplementedError


def IndTtest(s1, s2):
    '''独立样本T检验'''
    raise NotImplementedError
    
	
def RelTtest(s1, s2):
    '''配对样本T检验'''
    raise NotImplementedError
    
    
def ANOVA_OneWay(df, col_val, col_group):
    '''单因素方差分析，col_val为待比较列，col_group为组别列'''
    raise NotImplementedError
    